"""Schema validation for ArpTableEntry."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import ipaddress
from jsonschema import validate, FormatChecker

from dcim.models import Site
from ipam.models import Prefix, VRF

from netdoc import utils


def get_schema():
    """Return the JSON schema to validate Prefix data."""
    return {
        "type": "object",
        "properties": {
            "prefix": {
                "type": "string",
            },
            "vrf_id": {
                "type": "integer",
                "enum": list(VRF.objects.all().values_list("id", flat=True)),
            },
            "site_id": {
                "type": "integer",
                "enum": list(Site.objects.all().values_list("id", flat=True)),
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new Prefix objects."""
    schema = get_schema()
    schema["required"] = [
        "prefix",
    ]
    return schema


def address_to_network(address):
    """Convert an IP address with mask into the network address with mask."""
    try:
        # It's a network
        return str(ipaddress.ip_network(address))
    except ValueError:
        # It's a host
        try:
            return str(ipaddress.ip_interface(address).network)
        except ValueError as exc:
            raise ValueError(f"Invalid IP address {address}") from exc


def create(prefix=None, **kwargs):
    """Create an Prefix."""
    prefix = address_to_network(prefix)
    data = {
        "prefix": prefix,
        **kwargs,
    }
    data = utils.delete_empty_keys(data)
    validate(data, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(Prefix, **data)
    return obj


def get(prefix):
    """Return an Prefix."""
    prefix = address_to_network(prefix)
    obj = utils.object_get_or_none(Prefix, prefix=prefix)
    return obj



"""Schema validation for Device."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from dcim.models import (
    DeviceRole as DeviceRole_model,
    DeviceType as DeviceType_model,
    Manufacturer as Manufacturer_model,
    Site,
    Device,
)

from netdoc import utils
from netdoc.schemas import manufacturer as manufacturer_api, devicerole, devicetype


def get_schema():
    """Return the JSON schema to validate Device data."""
    return {
        "type": "object",
        "properties": {
            "name": {
                "type": "string",
                "transform": ["toUpperCase"],
            },
            "device_role_id": {
                "type": "integer",
                "enum": list(
                    DeviceRole_model.objects.all().values_list("id", flat=True)
                ),
            },
            "manufacturer_id": {
                "type": "integer",
                "enum": list(
                    Manufacturer_model.objects.all().values_list("id", flat=True)
                ),
            },
            "device_type_id": {
                "type": "integer",
                "enum": list(
                    DeviceType_model.objects.all().values_list("id", flat=True)
                ),
            },
            "serial_number": {
                "type": "string",
                "transform": ["toUpperCase"],
            },
            "site_id": {
                "type": "integer",
                "enum": list(Site.objects.all().values_list("id", flat=True)),
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new Device objects."""
    schema = get_schema()
    schema["required"] = [
        "name",
        "device_role_id",
        "device_type_id",
        "site_id",
    ]
    return schema


def create(manufacturer="Unknown", **kwargs):
    """Create a Device.

    Before need to get or create Manufacturer, DeviceModel, and DeviceType.
    """
    manufacturer_o = manufacturer_api.get(name=manufacturer)
    if not manufacturer_o:
        manufacturer_o = manufacturer_api.create(name=manufacturer)

    devicerole_o = devicerole.get(name="Unknown")
    if not devicerole_o:
        devicerole_o = devicerole.create(name="Unknown")

    devicetype_o = devicetype.get(model="Unknown", manufacturer_id=manufacturer_o.id)
    if not devicetype_o:
        devicetype_o = devicetype.create(
            model="Unknown", manufacturer_id=manufacturer_o.id
        )

    kwargs.update(
        {
            "device_role_id": devicerole_o.id,
            "device_type_id": devicetype_o.id,
        }
    )

    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(Device, **kwargs)
    return obj


def get(name):
    """Return a Device."""
    obj = utils.object_get_or_none(Device, name=name)
    return obj


def get_list(**kwargs):
    """Get a list of Device objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(Device, **kwargs)
    return result


def update(obj, **kwargs):
    """Update a Device."""
    update_if_not_set = ["serial_number"]

    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    kwargs_if_not_set = utils.filter_keys(kwargs, update_if_not_set)
    obj = utils.object_update(obj, **kwargs_if_not_set, force=False)
    return obj


def update_management(obj, discoverable_ip_address):
    """Update primary IP address if match the Discoverable IP address.

    Return True if management IP is set.
    """
    # Set management IP address
    for interface in obj.interfaces.filter(ip_addresses__isnull=False):
        # For each interface
        for ip_address_o in interface.ip_addresses.all():
            # For each configured IP address
            if discoverable_ip_address == str(ip_address_o.address.ip):
                obj.primary_ip4 = ip_address_o
                obj.save()
                return True
    return False



"""Useful functions."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import re
import random
import importlib
import ipaddress
import string
import json
import uuid
import macaddress

from nornir_netmiko.tasks import netmiko_send_command
from netmiko.utilities import get_structured_data
from textfsm.parser import TextFSMError

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User
from django.db.models import Q

from extras.scripts import get_scripts, run_script
from extras.models import JobResult, Script
from utilities.utils import NetBoxFakeRequest
from dcim.models import Interface


PLUGIN_SETTINGS = settings.PLUGINS_CONFIG.get("netdoc", {})
CONFIG_COMMANDS = [
    r"running-config",
    r"current-configuration",
]

FAILURE_OUTPUT = [
    # WARNING: must specify ^/\Z or a very unique string not found in valid outputs
    r"^$",  # Empty result
    r"^\s*Traceback",  # Python: traceback
    r"^\s*% Ambiguous command",  # Cisco: ambiguous command (command not supported)
    r"^\s*% Incomplete command",  # Cisco: incomplete command (command not supported)
    r"^\s*No link aggregation group exists",  # HP Comware: no port-channel configured
    r"^\s*% Unrecognized command found",  # HP Comware
    r"% \w+ is not enabled\s*\Z",  # Cisco: feature not enabled
    r"% \w+ not active\s*\Z",  # Cisco: feature not enabled
    r"% Wrong parameter found at",  # HP Comware
    r"% Unrecognized command found at",  # HP Comware
    r"% Invalid input detected at '\^' marker.\s*\Z",  # Cisco: invalid input detected
    r"% Invalid command at '\^' marker.\s*\Z",  # Cisco: invalid command detected
    r"No spanning tree instance exists.\s*\Z",  # Cisco STP
    r"No spanning tree instances exist.\s*\Z",  # Cisco STP
    r"Group\s+Port-channel\s+Protocol\s+Ports\s+[-+]+\s*\Z",  # Cisco etherchannel
    r"Group\s+Port-channel\s+Protocol\s+Ports\s+[-+]+\s*RU - L3",  # Cisco etherchannel
    r"Address\s+Age\s+MAC Address\s+Interface\s+Flags\s*\Z",  # Cisco ARP
    r"No VRF has been configured\s*\Z",  # Linux VRF
]

DRAWIO_ROLE_MAP = {
    "access-switch": {
        "style": "sketch=0;points=[[0.015,0.015,0],[0.985,0.015,0],[0.985,0.985,0],[0.015,0.985,0],"
        + "[0.25,0,0],[0.5,0,0],[0.75,0,0],[1,0.25,0],[1,0.5,0],[1,0.75,0],[0.75,1,0],[0.5,1,0],"
        + "[0.25,1,0],[0,0.75,0],[0,0.5,0],[0,0.25,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;shape=mxgraph.cisco19.rect;"
        + "prIcon=l2_switch;fillColor=#FAFAFA;strokeColor=#005073;",
        "width": 50,
        "height": 50,
    },
    "distribution-switch": {
        "style": "sketch=0;points=[[0.015,0.015,0],[0.985,0.015,0],[0.985,0.985,0],[0.015,0.985,0],"
        + "[0.25,0,0],[0.5,0,0],[0.75,0,0],[1,0.25,0],[1,0.5,0],[1,0.75,0],[0.75,1,0],[0.5,1,0],"
        + "[0.25,1,0],[0,0.75,0],[0,0.5,0],[0,0.25,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;shape=mxgraph.cisco19.rect;"
        + "prIcon=l3_switch;fillColor=#FAFAFA;strokeColor=#005073;",
        "width": 50,
        "height": 50,
    },
    "core-switch": {
        "style": "sketch=0;points=[[0.015,0.015,0],[0.985,0.015,0],[0.985,0.985,0],[0.015,0.985,0],"
        + "[0.25,0,0],[0.5,0,0],[0.75,0,0],[1,0.25,0],[1,0.5,0],[1,0.75,0],[0.75,1,0],[0.5,1,0],"
        + "[0.25,1,0],[0,0.75,0],[0,0.5,0],[0,0.25,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;shape=mxgraph.cisco19.rect;"
        + "prIcon=l3_modular;fillColor=#FAFAFA;strokeColor=#005073;",
        "width": 50,
        "height": 73,
    },
    "router": {
        "style": "sketch=0;points=[[0.5,0,0],[1,0.5,0],[0.5,1,0],[0,0.5,0],[0.145,0.145,0],"
        + "[0.8555,0.145,0],[0.855,0.8555,0],[0.145,0.855,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;shape=mxgraph.cisco19.rect;"
        + "prIcon=router;fillColor=#FAFAFA;strokeColor=#005073;",
        "width": 50,
        "height": 50,
    },
    "firewall": {
        "style": "sketch=0;points=[[0.015,0.015,0],[0.985,0.015,0],[0.985,0.985,0],[0.015,0.985,0],"
        + "[0.25,0,0],[0.5,0,0],[0.75,0,0],[1,0.25,0],[1,0.5,0],[1,0.75,0],[0.75,1,0],[0.5,1,0],"
        + "[0.25,1,0],[0,0.75,0],[0,0.5,0],[0,0.25,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;shape=mxgraph.cisco19.rect;"
        + "prIcon=firewall;fillColor=#FAFAFA;strokeColor=#005073;",
        "width": 64,
        "height": 50,
    },
    "server": {
        "style": "sketch=0;points=[[0.015,0.015,0],[0.985,0.015,0],[0.985,0.985,0],[0.015,0.985,0],"
        + "[0.25,0,0],[0.5,0,0],[0.75,0,0],[1,0.25,0],[1,0.5,0],[1,0.75,0],[0.75,1,0],[0.5,1,0],"
        + "[0.25,1,0],[0,0.75,0],[0,0.5,0],[0,0.25,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;"
        + "shape=mxgraph.cisco19.server;fillColor=#005073;strokeColor=none;",
        "width": 50,
        "height": 50,
    },
    "unknown": {
        "style": "sketch=0;points=[[0.015,0.015,0],[0.985,0.015,0],[0.985,0.985,0],[0.015,0.985,0],"
        + "[0.25,0,0],[0.5,0,0],[0.75,0,0],[1,0.25,0],[1,0.5,0],[1,0.75,0],[0.75,1,0],[0.5,1,0],"
        + "[0.25,1,0],[0,0.75,0],[0,0.5,0],[0,0.25,0]];verticalLabelPosition=bottom;html=1;"
        + "verticalAlign=top;aspect=fixed;align=center;pointerEvents=1;"
        + "shape=mxgraph.cisco19.server;fillColor=#005073;strokeColor=none;",
        "width": 27,
        "height": 50,
    },
}


def append_nornir_netmiko_task(
    task, commands, enable=True, template=None, order=128, supported=True
):
    """Append a Nornir task within a multiple_tasks adding extended details.

    commands can be str or list. template is allowed only if commands is str.
    """
    if isinstance(commands, str):
        commands = [commands]
    elif isinstance(commands, list) and template:
        raise ValueError("Cannot specify template for a list of commands")
    elif not isinstance(commands, list):
        raise ValueError("Command must be a string or a list of string")
    for command in commands:
        details = {
            "command": command,
            "template": template if template else command,
            "enable": enable,
            "order": order,
            "supported": supported,
        }
        for cmd_filter in PLUGIN_SETTINGS.get("NORNIR_SKIP_LIST"):
            if re.match(cmd_filter, command):
                # Skip excluded commands
                break
        else:
            # The filter does not match (filter loop completed successfully)
            task.run(
                task=netmiko_send_command,
                name=json.dumps(details),
                command_string=details.get("command"),
                use_textfsm=False,
                enable=details.get("enable"),
                use_timing=False,
                read_timeout=PLUGIN_SETTINGS.get("NORNIR_TIMEOUT"),
            )


def count_interface_neighbors(neighbors_list, key):
    """Given a Netmiko parsed output, count CDP/LLDP neighbors per interface."""
    neighbors_per_interface = {}
    for item in neighbors_list:
        label = normalize_interface_label(item.get(key))
        if label not in neighbors_per_interface:
            neighbors_per_interface[label] = 0
        neighbors_per_interface[label] += 1
    return neighbors_per_interface


def delete_empty_keys(dct):
    """Delete None values recursively from a dictionary."""
    for key, value in list(dct.items()):
        if isinstance(value, dict):
            delete_empty_keys(value)
        elif value is None:
            del dct[key]
        elif isinstance(value, list):
            if not value:
                del dct[key]
            else:
                for v_i in value:
                    if isinstance(v_i, dict):
                        delete_empty_keys(v_i)
    return dct


def export_log(log):
    """Export DiscoveryLog in JSON format."""
    data = {
        "command": log.command,
        "configuration": log.configuration,
        "details": log.details,
        "order": log.order,
        "parsed_output": log.parsed_output,
        "raw_output": log.raw_output,
        "template": log.template,
        "success": log.success,
        "parsed": log.parsed,
        "ingested": log.ingested,
        "discoverable__mode": log.discoverable.mode,
    }
    return data


def filter_keys(dct, allowed_keys):
    """Return a dict with only allowed keys."""
    filtered_dct = {}
    for key, value in dct.items():
        if key in allowed_keys:
            filtered_dct[key] = value
    return filtered_dct


def get_random_string(length):
    """Create a random lowercase string."""
    letters = string.ascii_lowercase
    result_str = "".join(random.choice(letters) for i in range(length))  # nosec
    return result_str


def get_diagram_interfaces(
    mode, sites=None, roles=None, vrfs=None, include_global_vrf=False
):
    """Get relevant Interface queryset according to Diagram settings."""
    if not sites:
        sites = []
    if not roles:
        roles = []
    if not vrfs:
        vrfs = []

    interface_qs = Interface.objects.all()
    if sites:
        interface_qs = interface_qs.filter(device__site_id__in=sites)
    if roles:
        interface_qs = interface_qs.filter(device__device_role_id__in=roles)
    if mode == "l2":
        # Filter out Interface without cable
        interface_qs = interface_qs.exclude(cable__isnull=True)
        if vrfs:
            interface_qs = interface_qs.filter(ip_addresses__vrf_id__in=vrfs)
    if mode == "l3":
        # Filter out Interface without IP address
        interface_qs = interface_qs.exclude(ip_addresses__isnull=True)
        if include_global_vrf and not vrfs:
            # Include Global VRF only
            interface_qs = interface_qs.filter(ip_addresses__vrf__isnull=True)
        elif include_global_vrf and vrfs:
            # Include selected VRFs and Global VRF
            interface_qs = interface_qs.filter(
                Q(ip_addresses__vrf__isnull=True) | Q(ip_addresses__vrf_id__in=vrfs)
            )

        elif not include_global_vrf and vrfs:
            # Include selected VRFs only
            interface_qs = interface_qs.filter(ip_addresses__vrf_id__in=vrfs)

    return interface_qs


def get_remote_lldp_interface_label(
    port_id="", port_description="", system_description=""
):  # pylint: disable=unused-argument
    """
    Return the remote interface label.

    None is returned if no valid interface is found.
    The interface is selected based on port_id, port_description and system_description.
    The function tries to detect if port_id or port_description contain the interface description.
    """
    try:
        normalize_mac_address(port_id)
        # port_id contains a MAC address
        port_id = None
    except ValueError:
        pass

    try:
        normalize_mac_address(port_description)
        # port_description contains a MAC address
        port_description = None
    except ValueError:
        pass

    if port_id and re.match(r"\*\*\*|.*###|.*\[.*\]", port_id):
        # port_id contains a description
        port_id = None

    if port_description and re.match(r"\*\*\*|.*###|.*\[.*\]", port_description):
        # port_description contains a description
        port_description = None

    if port_id:
        # Give prevedence to port_id
        return normalize_interface_label(port_id)

    # Use port_description as last resort
    return normalize_interface_label(port_description)


def incomplete_mac(mac):
    """Return True if the MAC address is incomplete (from ARP table)."""
    if not mac:
        return True
    if "incomplete" in mac.lower():
        return True
    return False


def is_hostname(name):
    """Return true if the hostname is valid."""
    if not name:
        # Hostname is empty
        return False
    allowed = re.compile(r"(?!-)[A-Z_\d-]{1,63}(?<!-)$", re.IGNORECASE)
    if all(allowed.match(x) for x in name.split(".")):
        return True
    # Hostname contains invalid chars
    return False


def log_ingest(log):
    """Ingest a log calling the custom ingestor."""
    function_name = f"{log.discoverable.mode}_{log.template}"
    function_name = function_name.replace(" ", "_")
    function_name = function_name.replace("-", "_")
    function_name = function_name.lower()

    try:
        module = importlib.import_module(f"netdoc.ingestors.{function_name}")
    except ModuleNotFoundError as exc:
        raise ModuleNotFoundError(f"Ingestor not found for {function_name}") from exc

    if log.order > 0 and not log.discoverable.device:
        # Log with order > 0 must have a Device attached to the parent Discoverable
        raise ValueError(
            f"The discoverable {log.discoverable.address} does not have an attached device "
            + "thus logs cannot be ingested. Check if logs with priority 0 are ingested or if "
            + "Device is attached to a different Discoverable."
        )

    module.ingest(log)
    return log


def normalize_hostname(name):
    """Hostnames must be uppercase without domain."""
    if name:
        try:
            # Check if it's a valid IP Address
            ipaddress.ip_address(name)
            return name
        except ValueError:
            # Not a valid IP Address, it is a hostname
            pass
        name = name.split(".")[0]  # Removes domain if exists
        name = name.upper()
        name = re.sub(
            r"\(.+?\)", "", name
        )  # NX-OS: remove serial number (e.g. hostname(serialnumber))
    return name


def normalize_interface_duplex(duplex):
    """Normalize Interface.duplex."""
    if not duplex:
        # None is allowed
        return None
    duplex = duplex.lower()
    if "auto" in duplex:
        return "auto"
    if "half" in duplex:
        return "half"
    if "full" in duplex:
        return "full"
    if "unknown" in duplex:
        # Unknown
        return None
    raise ValueError(f"Invalid duplex mode {duplex}")


def normalize_interface_label(name):
    """Given an interface name, return the shortname (label)."""
    name = name.lower()
    if name.startswith("gigabitethernet"):
        return name.replace("gigabitethernet", "gi")
    if name.startswith("fastethernet"):
        return name.replace("fastethernet", "fa")
    if name.startswith("tengigabitethernet"):
        return name.replace("tengigabitethernet", "te")
    if name.startswith("ten-gigabitethernet"):  # HP Comware
        return name.replace("ten-gigabitethernet", "te")
    if name.startswith("m-gigabitethernet"):  # HP Comware
        return name.replace("m-gigabitethernet", "mge")
    if name.startswith("xge"):  # HP Comware
        return name.replace("xge", "te")
    if name.startswith("fortygige"):
        return name.replace("fortygige", "fge")
    if name.startswith("hundredgige"):
        return name.replace("hundredgige", "hge")
    if name.startswith("ethernet"):
        return name.replace("ethernet", "e")
    if name.startswith("eth"):
        return name.replace("eth", "e")
    if name.startswith("et"):
        return name.replace("et", "e")
    if name.startswith("vlan-interface"):  # HP Comware
        return name.replace("vlan-interface", "vl")
    if name.startswith("vlan"):
        return name.replace("vlan", "vl")
    if name.startswith("management"):
        return name.replace("management", "mgmt")
    if name.startswith("mgmteth0"):
        # Cisco XR Management interface
        return name.replace("mgmteth0", "mg0")
    if name.startswith("loopback"):
        return name.replace("loopback", "lo")
    if name.startswith("port-channel"):
        return name.replace("port-channel", "po")
    if name.startswith("route-aggregation"):
        return name.replace("route-aggregation", "ragg")
    if name.startswith("bridge-aggregation"):
        return name.replace("bridge-aggregation", "bagg")
    if name.startswith("tunnel"):
        return name.replace("tunnel", "tu")
    if name.startswith(
        "ge"
    ):  # HP Comware is using "gi" for GigabitEthernet, while Cisco is using "gi"
        return name.replace("ge", "gi")
    return name


def normalize_interface_mode(mode):
    """Normalize Interface mode (access/trunk/...)."""
    if not mode:
        # None is allowed
        return None
    mode = mode.lower()
    if "trunk" in mode:
        return "tagged"
    if "hybrid" in mode:
        return "tagged"
    if "access" in mode:
        return "access"
    if mode == "fex-fabric":
        # Cisco FEX
        return "tagged-all"
    if "private-vlan" in mode:
        # PVLAN still not supported by NetBox
        return None
    if "dot1q-tunnel" in mode:
        # QinQ still not supported by Netbox
        return None
    if "down" in mode:
        # Ignore down interfaces
        return None
    raise ValueError(f"Invalid interface mode {mode}")


def normalize_interface_mtu(mtu):
    """
    Normalize MTU value returning an Interger.

    None is returned if mtu is empty or None.
    """
    if not mtu:
        # Return None if value is empty or None
        return None
    try:
        return int(mtu)
    except ValueError as exc:
        raise ValueError(f"Invalid MTU {mtu}") from exc


def normalize_interface_speed(speed):
    """Normalize Interface speed in kb/s."""
    if not speed:
        # None is allowed
        return None

    speed = speed.lower()
    if "auto" in speed:
        # Speed is set to auto
        return None
    speed = speed.replace(" ", "")
    speed = speed.replace("kbit", "")
    speed = speed.replace("kbps", "")
    speed = speed.replace("mbps", "000")
    speed = speed.replace("mb/s", "000")
    speed = speed.replace("gb/s", "000000")
    speed = speed.replace("gbps", "000000")
    try:
        speed = int(speed)
    except ValueError as exc:
        raise ValueError(f"Invalid interface speed {speed}") from exc
    return speed


def normalize_interface_status(status):
    """Normalize interface status.

    Return True if the link is up, False elsewhere.
    """
    status = status.lower()
    if "up" in status:
        return True
    if "down" in status:
        return False
    raise ValueError(f"Invalid interface status {status}")


def normalize_interface_type(name="", encapsulation=""):
    """Return interface type from name/encapsulation."""
    label = normalize_interface_label(name)
    encapsulation = encapsulation.lower()
    if parent_interface(label):
        # Subinterface
        return "virtual"
    if re.match(r"^(po|bond).*", label):
        # LAG (portchannel, bond)
        return "lag"
    if re.match(r"lo.*", label):
        # Loopback
        return "virtual"
    if re.match(r"tun.*", label):
        # Tunnel
        return "virtual"
    if re.match(r"vl.*", label):
        # SVI (VLAN interface)
        return "bridge"
    if re.match(r"null.*", label):
        # Null
        return "virtual"
    if "loopback" in encapsulation:
        # Loopback
        return "virtual"
    if "tunnel" in encapsulation:
        # Tunnel
        return "virtual"
    # If not found return "other"
    return "other"


def normalize_mac_address(mac_address):
    """Return a MAC Address in the format 01:23:45:67:89:AB."""
    if not mac_address:
        # Convert None in empty string to raise ValueError only
        mac_address = ""

    # Removing non hex chars
    mac_address = mac_address.replace("-", "")
    mac_address = mac_address.replace(":", "")
    # Converting into a MAC object (as a validator)
    mac_address_o = macaddress.MAC(mac_address)
    # Return the MAC address in Netbox format
    return str(mac_address_o).replace("-", ":")


def normalize_ip_address_or_none(ip_address):
    """
    Return IP address (10.1.2.3) or None from string.

    String can be 10.1.2.3 or 10.1.2.3/24.
    """
    try:
        # 10.1.2.3
        ip_address_o = ipaddress.ip_address(ip_address)
        return str(ip_address_o)
    except ValueError:
        pass

    try:
        # 10.1.2.3/24
        ip_network_o = ipaddress.ip_interface(ip_address)
        return str(ip_network_o.ip)
    except ValueError:
        pass

    return None


def normalize_route_type(route_type):
    """Return route type protocol."""
    route_type = route_type.lower()
    if route_type in ["c", "direct", "local", "hsrp", "l"]:
        # Connected
        return "c"
    if route_type in ["s", "static", "s*"]:
        # Static
        return "s"
    if route_type in ["r", "rip-10"]:
        # RIP
        return "r"
    if route_type in ["b", "bgp"]:
        # BGP
        return "b"
    if route_type in ["d"]:
        # EIGRP
        return "e"
    if route_type in ["ex"]:
        # EIGRP External
        return "ex"
    if route_type in ["o", "ospf", "o_intra", "o ia"]:
        # OSPF Inter Area
        return "oia"
    if route_type in ["n1", "o n1"]:
        # OSPF NSSA Type 1
        return "on1"
    if route_type in ["n2", "o n2"]:
        # OSPF NSSA Type 2
        return "on2"
    if route_type in ["e1", "o e1", "o_ase1"]:
        # OSPF External Type 1
        return "oe1"
    if route_type in ["e2", "o e2", "o_ase2", "o_ase"]:
        # OSPF External Type 2
        return "oe2"
    if route_type in ["i"]:
        # IS-IS
        return "i"
    if route_type in ["su"]:
        # IS-IS Summary
        return "is"
    if route_type in ["l1"]:
        # IS-IS L1
        return "i1"
    if route_type in ["l2"]:
        # IS-IS
        return "i2"
    if route_type in ["unknown"]:
        # Unknown
        return "u"
    raise ValueError(f"Invalid route type {route_type}")


def normalize_serial(serial):
    """Serial number must be uppercase."""
    if serial:
        serial = serial.upper()
    return serial


def normalize_vlan_list(trunking_vlans):
    """Normalize a VLAN list (can be a comma separated string or a list)."""
    vlans = []
    if isinstance(trunking_vlans, list):
        for vlan in trunking_vlans:
            vlans.extend(normalize_vlan_range(vlan))
    else:
        vlans.extend(normalize_vlan_range(trunking_vlans))
    return vlans


def normalize_vlan_range(vlan):
    """Normalize a single VLAN or VLAN range, returning a list of integer."""
    if isinstance(vlan, int):
        # Integer
        return [vlan]

    vlan = vlan.lower()
    vlan = vlan.replace("(default vlan)", "")
    vlan = vlan.replace(" ", "")
    if vlan == "none":
        return []
    if vlan == "all":
        # All VLANs
        return normalize_vlan_range("1-4094")
    if "," in vlan:
        # Set of VLANs (must before VLAN range)
        vlans = []
        for value in vlan.split(","):
            vlans.extend(normalize_vlan_range(value))
        return vlans
    if "-" in vlan:
        # VLAN range
        try:
            return list(range(int(vlan.split("-")[0]), int(vlan.split("-")[1]) + 1))
        except ValueError as exc:
            raise ValueError(f"cannot convert VLAN {vlan} range to integer") from exc
    try:
        return [int(vlan)]
    except ValueError as exc:
        raise ValueError(f"cannot convert VLAN {vlan} to integer") from exc


def object_create(model_o, **kwargs):
    """Create a generic Model object using kwargs."""
    return model_o.objects.create(**kwargs)


def object_get_or_none(model_o, **kwargs):
    """Get a generic Model object using kwargs. None is returned if not found."""
    try:
        return model_o.objects.get(**kwargs)
    except model_o.DoesNotExist:
        return None


def object_list(model_o, **kwargs):
    """Get a list of Model objects."""
    result = model_o.objects.filter(**kwargs)
    return list(result)


def object_update(obj, force=True, **kwargs):
    """Update a generic Model object using kwargs. Object is saved only if modified.

    With force=False, the field is updated only if not set.
    """
    updated = False
    for key, new_value in kwargs.items():
        current_value = getattr(obj, key)
        if new_value == current_value:
            # No need to update
            continue
        if force:
            # Always update
            setattr(obj, key, new_value)
            updated = True
        if current_value is None:
            # Update only if not none
            setattr(obj, key, new_value)
            updated = True

    if updated:
        obj.save()
    return obj


def parent_interface(label):
    """If subinterface return parent interface else return None."""
    label = normalize_interface_label(label)
    if re.match(r"^[^.]+\.[0-9]+$", label):
        # Contains only one "." and ends with numbers
        parent_label = re.sub(r".[0-9]+$", "", label)
        return parent_label
    return None


def parse_netmiko_output(output, command, platform, template=None):
    """Parse Netmiko output using NTC templates."""
    if not template:
        # If template is empty, use command as template
        template = command
    if template == "HOSTNAME":
        # Parsed during ingestion
        return output, True
    try:
        parsed_output = get_structured_data(output, platform=platform, command=template)
        if isinstance(parsed_output, str) and parsed_output == output:
            # NTC template not found
            return (
                "Output not parsed, check if NTC template exists and debug it.",
                False,
            )
        if not isinstance(parsed_output, dict) and not isinstance(parsed_output, list):
            parsed_output = f"Not a valid dict or list\n\n{parsed_output}"
            return parsed_output, False
        return parsed_output, True
    except TextFSMError as exc:
        return str(exc), False


def spawn_script(script_name, get_data=None, post_data=None, file_list=None, user=None):
    """Spawn a Netbox Script defined into scripts folder."""
    if not get_data:
        get_data = {}
    if not post_data:
        post_data = {}
    if not file_list:
        file_list = {}
    if not user:
        user = User.objects.filter(is_superuser=True).order_by("pk")[0]

    script = get_scripts().get("NetDoc").get(script_name)
    request = NetBoxFakeRequest(
        {
            "META": {},
            "POST": post_data,
            "GET": get_data,
            "FILES": file_list,
            "user": user,
            "id": uuid.uuid4(),
        }
    )

    JobResult.enqueue_job(
        run_script,
        name=script.full_name,
        obj_type=ContentType.objects.get_for_model(Script),
        user=request.user,  # pylint: disable=no-member
        schedule_at=None,
        interval=None,
        job_timeout=script.job_timeout,
        data=request.POST,  # pylint: disable=no-member
        request=request,
    )



"""Ingestor for netmiko_hp_comware_display_vlan_all."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import vlan


def ingest(log):
    """Processing parsed output.

    VLAN - Interface association is ingested in the "sdisplay interface" output.
    """
    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/hp_comware/display_vlan_all # pylint: disable=line-too-long
        vlan_id = int(item.get("vlan_id"))
        vlan_name = item.get("name")

        vlan_o = vlan.get(vlan_id, vlan_name)
        if not vlan_o:
            data = {
                "name": vlan_name,
                "vid": vlan_id,
            }
            vlan_o = vlan.create(**data)

    # Update the log
    log.ingested = True
    log.save()



"""Schema validation for RouteTableEntry."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from dcim.models import Interface, Device
from ipam.models import VRF

from netdoc.models import RouteTableEntry, RouteTypeChoices
from netdoc import utils


def get_schema():
    """Return the JSON schema to validate RouteTableEntry data."""
    return {
        "type": "object",
        "properties": {
            "device_id": {
                "type": "integer",
                "enum": list(Device.objects.all().values_list("id", flat=True)),
            },
            "destination": {
                "type": "string",  # IP Network
            },
            "nexthop_ip": {
                "type": "string",  # IP Address
            },
            "nexthop_if_id": {
                "type": "integer",
                "enum": list(Interface.objects.all().values_list("id", flat=True)),
            },
            "vrf_id": {
                "type": "integer",
                "enum": list(VRF.objects.all().values_list("id", flat=True)),
            },
            "distance": {
                "type": "integer",
            },
            "metric": {
                "type": "integer",
            },
            "protocol": {
                "type": "string",
                "enum": [key for key, value in RouteTypeChoices()],
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new RouteTableEntry objects."""
    schema = get_schema()
    schema["required"] = [
        "device_id",
        "destination",
        "protocol",
    ]
    return schema


def create(**kwargs):
    """Create an RouteTableEntry."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    # At least one of nexthop_ip or nexthop_if should be set
    if not kwargs.get("nexthop_ip") and not kwargs.get("nexthop_if_id"):
        raise ValueError("Even one of nexthop_ip or nexthop_if_id should have a value.")

    obj = utils.object_create(RouteTableEntry, **kwargs)
    return obj


def get(
    device_id=None,
    destination=None,
    distance=None,
    metric=None,
    protocol=None,
    discovered=True,
):
    """Return an RouteTableEntry."""
    obj = utils.object_get_or_none(
        RouteTableEntry,
        device_id=device_id,
        destination=destination,
        distance=distance,
        metric=metric,
        protocol=protocol,
    )
    if obj and discovered:
        # Update updated_at
        obj.save()
    return obj


def get_list(**kwargs):
    """Get a list of RouteTableEntry objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(RouteTableEntry, **kwargs)
    return result



"""
NetDoc PYPI setup file.

Install with: python3 setup.py install
Develop with: python3 setup.py develop
Make it available on PIP with:
    python3 setup.py sdist
    pip3 install twine
    twine upload dist/*
"""

__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"
__version__ = "0.9.54"

from setuptools import find_packages, setup

setup(
    name="netdoc",
    version=__version__,
    description="Network Documentation plugin for NetBox",
    url="https://github.com/dainok/netdoc",
    author="Andrea Dainese",
    author_email="andrea@adainese.it",
    license="GNU v3.0",
    install_requires=[
        "jsonschema",
        "python-slugify",
        "netmiko==4.1.2",
        "nornir==3.3.0",
        "nornir_netmiko==0.2.0",
        "nornir_utils",
        "ipaddress",
        "macaddress",
        "ouilookup",
        "n2g==0.3.3",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
    ],
    project_urls={
        "Source": "https://github.com/dainok/netdoc",
    },
)



# Generated by Django 4.1.6 on 2023-02-14 12:16

from django.db import migrations, models
import taggit.managers
import utilities.json


class Migration(migrations.Migration):
    dependencies = [
        ("ipam", "0063_standardize_description_comments"),
        ("dcim", "0167_module_status"),
        ("extras", "0084_staging"),
        (
            "netdoc",
            "0003_alter_arptableentry_options_alter_credential_options_and_more",
        ),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="macaddresstableentry",
            options={
                "ordering": ["mac_address", "vvid"],
                "verbose_name": "MAC Address table entry",
                "verbose_name_plural": "MAC Address table entries",
            },
        ),
        migrations.AlterModelOptions(
            name="routetableentry",
            options={
                "ordering": ["device", "protocol", "metric"],
                "verbose_name": "Route",
                "verbose_name_plural": "Routes",
            },
        ),
        migrations.CreateModel(
            name="Diagram",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("details", models.JSONField(default=dict, editable=False)),
                ("mode", models.CharField(max_length=30)),
                ("name", models.CharField(max_length=100)),
                (
                    "device_roles",
                    models.ManyToManyField(
                        blank=True, related_name="+", to="dcim.devicerole"
                    ),
                ),
                (
                    "sites",
                    models.ManyToManyField(
                        blank=True, related_name="+", to="dcim.site"
                    ),
                ),
                (
                    "tags",
                    taggit.managers.TaggableManager(
                        through="extras.TaggedItem", to="extras.Tag"
                    ),
                ),
                (
                    "vrfs",
                    models.ManyToManyField(blank=True, related_name="+", to="ipam.vrf"),
                ),
            ],
            options={
                "verbose_name": "Diagram",
                "verbose_name_plural": "Diagrams",
                "ordering": ["name"],
                "unique_together": {("name",)},
            },
        ),
    ]



# Generated by Django 4.1.6 on 2023-02-12 09:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("dcim", "0167_module_status"),
        ("ipam", "0063_standardize_description_comments"),
        ("netdoc", "0002_alter_arptableentry_custom_field_data_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="arptableentry",
            options={
                "ordering": ["ip_address"],
                "verbose_name": "ARP table entry",
                "verbose_name_plural": "ARP table entries",
            },
        ),
        migrations.AlterModelOptions(
            name="credential",
            options={
                "ordering": ["name"],
                "verbose_name": "Credential",
                "verbose_name_plural": "Credentials",
            },
        ),
        migrations.AlterModelOptions(
            name="discoverable",
            options={
                "ordering": ["mode", "address"],
                "verbose_name": "Device",
                "verbose_name_plural": "Devices",
            },
        ),
        migrations.AlterModelOptions(
            name="discoverylog",
            options={
                "ordering": ["created"],
                "verbose_name": "Log",
                "verbose_name_plural": "Logs",
            },
        ),
        migrations.AlterModelOptions(
            name="routetableentry",
            options={
                "ordering": ("device", "protocol", "metric"),
                "verbose_name": "Route ",
                "verbose_name_plural": "Routes",
            },
        ),
        migrations.RenameField(
            model_name="routetableentry",
            old_name="type",
            new_name="protocol",
        ),
        migrations.AlterUniqueTogether(
            name="routetableentry",
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name="discoverylog",
            name="request",
        ),
        migrations.AddField(
            model_name="discoverylog",
            name="details",
            field=models.JSONField(default=dict, editable=False),
        ),
        migrations.AddField(
            model_name="discoverylog",
            name="order",
            field=models.IntegerField(default=128),
        ),
        migrations.AddField(
            model_name="discoverylog",
            name="template",
            field=models.CharField(
                default="migrations", editable=False, max_length=255
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="discoverylog",
            name="raw_output",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="routetableentry",
            name="distance",
            field=models.IntegerField(blank=True, editable=False, null=True),
        ),
        migrations.AlterField(
            model_name="routetableentry",
            name="metric",
            field=models.IntegerField(blank=True, editable=False, null=True),
        ),
        migrations.AlterUniqueTogether(
            name="routetableentry",
            unique_together={
                ("device", "destination", "distance", "metric", "protocol", "vrf")
            },
        ),
    ]



"""Ingestor for linux_ip_vrf_show."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netdoc.schemas import vrf


def ingest(log):
    """Processing parsed output."""
    for item in log.parsed_output:
        # See https://github.com/networktocode/ntc-templates/tree/master/tests/linux/ip_vrf_show # pylint: disable=line-too-long
        vrf_name = item.get("vrf")

        vrf_o = vrf.get(name=vrf_name)
        if not vrf_o:
            data = {
                "name": vrf_name,
            }
            vrf_o = vrf.create(mandatory_rd=False, **data)

    # Update the log
    log.ingested = True
    log.save()



"""Forms, called by Views."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from django import forms

from dcim.models import Device, Site, DeviceRole
from ipam.models import VRF

from netbox.forms import (
    NetBoxModelForm,
    NetBoxModelFilterSetForm,
    NetBoxModelCSVForm,
    NetBoxModelBulkEditForm,
)
from utilities.forms import (
    CSVModelChoiceField,
    DynamicModelChoiceField,
    DynamicModelMultipleChoiceField,
    StaticSelect,
    BOOLEAN_WITH_BLANK_CHOICES,
    add_blank_choice,
)

from netdoc.models import (
    Credential,
    Discoverable,
    DiscoveryLog,
    DiscoveryModeChoices,
    DiagramModeChoices,
)

#
# Credential forms
#


class CredentialForm(NetBoxModelForm):
    """Form used to add/edit Credential."""

    username = forms.CharField(required=False)
    password = forms.CharField(required=False, widget=forms.PasswordInput)
    enable_password = forms.CharField(required=False, widget=forms.PasswordInput)

    class Meta:
        """Form metadata."""

        model = Credential
        fields = [
            "name",
            "username",
            "password",
            "enable_password",
            "tags",
        ]


class CredentialCSVForm(NetBoxModelCSVForm):
    """Form used to add Credential objects via CSV import."""

    class Meta:
        """Form metadata."""

        model = Credential
        fields = ["name", "username", "password", "enable_password"]


class CredentialBulkEditForm(NetBoxModelBulkEditForm):
    """Form used to bulk edit Credential objects."""

    username = forms.CharField(required=False)
    password = forms.CharField(required=False, widget=forms.PasswordInput)
    enable_password = forms.CharField(required=False, widget=forms.PasswordInput)

    model = Credential
    nullable_fields = ["username", "password", "enable_password"]


#
# Diagram forms
#


class DiagramForm(NetBoxModelForm):
    """Form used to add/edit Diagram."""

    name = forms.CharField(required=False)
    mode = forms.ChoiceField(choices=DiagramModeChoices, required=True)
    vrfs = DynamicModelMultipleChoiceField(
        queryset=VRF.objects.all(),
        required=False,
    )
    include_global_vrf = forms.BooleanField(
        required=False,
        initial=False,
        help_text="If set and no VRF is selected, Global only is included."
        + " If set and VRFs are selected, Global is included.",
    )
    sites = DynamicModelMultipleChoiceField(
        queryset=Site.objects.all(),
        required=False,
    )
    device_roles = DynamicModelMultipleChoiceField(
        queryset=DeviceRole.objects.all(),
        required=False,
    )

    class Meta:
        """Form metadata."""

        model = Credential
        fields = [
            "name",
            "mode",
            "device_roles",
            "sites",
            "vrfs",
            "include_global_vrf",
            "tags",
        ]


#
# Discoverable views
#


class DiscoverableForm(NetBoxModelForm):
    """Form used to add/edit Discoverable."""

    address = forms.GenericIPAddressField()
    credential = forms.ModelChoiceField(
        queryset=Credential.objects.all(),
        required=True,
    )
    mode = forms.ChoiceField(choices=DiscoveryModeChoices, required=True)
    device = DynamicModelChoiceField(queryset=Device.objects.all(), required=False)
    discoverable = forms.BooleanField(
        required=False,
        initial=True,
    )
    site = DynamicModelChoiceField(
        queryset=Site.objects.all(),
        help_text="Site",
        required=True,
    )

    class Meta:
        """Form metadata."""

        model = Discoverable
        fields = [
            "address",
            "device",
            "credential",
            "mode",
            "discoverable",
            "site",
            "tags",
        ]


class DiscoverableCSVForm(NetBoxModelCSVForm):
    """Form used to add Discoverable objects via CSV import."""

    address = forms.GenericIPAddressField(
        help_text="Management IP address",
    )
    credential = CSVModelChoiceField(
        queryset=Credential.objects.all(),
        required=True,
        to_field_name="name",
        help_text="Assigned credential",
    )
    mode = forms.ChoiceField(
        choices=DiscoveryModeChoices,
        required=True,
        help_text="Discovery mode",
    )
    site = CSVModelChoiceField(
        queryset=Site.objects.all(),
        to_field_name="name",
        help_text="Site",
        required=True,
    )

    class Meta:
        """Form metadata."""

        model = Discoverable
        fields = ["address", "credential", "mode", "site"]


class DiscoverableBulkEditForm(NetBoxModelBulkEditForm):
    """Form used to bulk edit Discoverable objects."""

    credential = CSVModelChoiceField(
        queryset=Credential.objects.all(),
        required=False,
        to_field_name="name",
        help_text="Assigned credential",
    )
    mode = forms.ChoiceField(
        choices=add_blank_choice(DiscoveryModeChoices),
        required=False,
        initial="",
        widget=StaticSelect(),
        help_text="Discovery mode",
    )
    discoverable = forms.NullBooleanField(
        help_text="Is discoverable?",
        required=False,
    )
    site = forms.ModelChoiceField(
        queryset=Site.objects.all(),
        to_field_name="name",
        help_text="Site",
        required=False,
    )

    model = Discoverable
    nullable_fields = "device"


class DiscoveryLogListFilterForm(NetBoxModelFilterSetForm):
    """Form used to filter DiscoveryLog using parameters. Used in DiscoveryLogListView."""

    model = DiscoveryLog
    configuration = forms.NullBooleanField(
        required=False,
        label="Configuration output",
        widget=StaticSelect(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )
    success = forms.NullBooleanField(
        required=False,
        label="Completed successfully",
        widget=StaticSelect(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )
    supported = forms.NullBooleanField(
        required=False,
        label="Command is supported",
        widget=StaticSelect(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )
    parsed = forms.NullBooleanField(
        required=False,
        label="Parsed successfully",
        widget=StaticSelect(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )
    ingested = forms.NullBooleanField(
        required=False,
        label="Ingested successfully",
        widget=StaticSelect(choices=BOOLEAN_WITH_BLANK_CHOICES),
    )



# Generated by Django 4.1.6 on 2023-02-21 07:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("netdoc", "0004_alter_macaddresstableentry_options_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="diagram",
            name="include_global_vrf",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="diagram",
            name="details",
            field=models.JSONField(default=dict),
        ),
    ]



"""Schema validation for ArpTableEntry."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from dcim.models import Interface as Interface_model

from netdoc.models import ArpTableEntry
from netdoc import utils


def get_schema():
    """Return the JSON schema to validate ArpTableEntry data."""
    return {
        "type": "object",
        "properties": {
            "interface_id": {
                "type": "integer",
                "enum": list(
                    Interface_model.objects.all().values_list("id", flat=True)
                ),
            },
            "ip_address": {
                "type": "string",
            },
            "mac_address": {
                "type": "string",
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new ArpTableEntry objects."""
    schema = get_schema()
    schema["required"] = [
        "interface_id",
        "ip_address",
        "mac_address",
    ]
    return schema


def create(**kwargs):
    """Create an ArpTableEntry."""
    kwargs = utils.delete_empty_keys(kwargs)
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())
    obj = utils.object_create(ArpTableEntry, **kwargs)
    return obj


def get(interface_id, ip_address, mac_address, discovered=True):
    """Return an ArpTableEntry."""
    obj = utils.object_get_or_none(
        ArpTableEntry,
        interface_id=interface_id,
        ip_address=ip_address,
        mac_address=mac_address,
    )
    if obj and discovered:
        # Update updated_at
        obj.save()
    return obj


def get_list(**kwargs):
    """Get a list of ArpTableEntry objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(ArpTableEntry, **kwargs)
    return result



"""Functions to build vis.js topology data from a queryset."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

import hashlib
import ipaddress
from N2G import drawio_diagram

from jinja2 import Template

from netdoc.models import DeviceImageChoices
from netdoc.utils import DRAWIO_ROLE_MAP

NODE_TEMPLATE = """
<table>
    <tbody>
        <tr>
            <th>Type: </th>
            <td>{{ device.device_type.model }}</td>
        </tr>
        {% if device.primary_ip %}
            <tr>
                <th>IP Address: </th>
                <td>{{ device.primary_ip }}</td>
            </tr>
        {% endif %}
        {% if device.site %}
            <tr>
                <th>Site: </th>
                <td>{{ device.site.name }}</td>
            </tr>
        {% endif %}
        {% if device.location %}
            <tr>
                <th>Location: </th>
                <td>{{ device.location.name }}</td>
            </tr>
        {% endif %}
    </tbody>
</table>
"""

CABLE_TEMPLATE = """
<table>
    <tbody>
        <tr>
            <th>From: </th>
            <td>{{ from_interface.device.name }}:{{ from_interface.name }}</td>
        </tr>
        <tr>
            <th>To: </th>
            <td>{{ to_interface.device.name }}:{{ to_interface.name }}</td>
        </tr>
    </tbody>
</table>
"""

NETWORK_TEMPLATE = """
<table>
    <tbody>
        {% if network %}
        <tr>
            <th>Network: </th>
            <td>{{ network.compressed }}</td>
        </tr>
        <tr>
            <th>VRF: </th>
            <td>{% if vrf %}{{ vrf.name }}{% else %}Global{% endif %}</td>
        </tr>
        {% endif %}
        {% if interface %}
        <tr>
            <th>Interface: </th>
            <td>{{ interface.name }}</td>
        </tr>
        <tr>
            <th>VRF: </th>
            <td>{% if interface.vrf %}{{ interface.vrf.name }}{% else %}Global{% endif %}</td>
        </tr>
        <tr>
            <th>Address: </th>
            <td>{{ address.address }}</td>
        </tr>
        {% endif %}
    </tbody>
</table>
"""


def integer_hash(text, length=16):
    """Return a very simple and insecure interger hash.

    Used in L3 diagrams to link networks with VRF aware devices.
    """
    return int(hashlib.sha256(text.encode("utf-8")).hexdigest(), 16) % 10**length


def get_l2_topology_data(queryset, details):
    """Create a L2 vis.js topology data from an Interface queryset."""
    nodes = {}
    links = {}

    for interface_o in queryset:
        # Create device
        device_o = interface_o.device
        device_id = device_o.pk
        if device_id not in nodes:
            device_role = device_o.device_role.slug
            if device_role in [key for key, value in DeviceImageChoices()]:
                image_url = f"/static/netdoc/img/{device_role}.png"
            else:
                image_url = "/static/netdoc/img/unknown.png"
            nodes[device_id] = {
                "id": device_id,
                "label": interface_o.device.name,
                "role": device_role,
                "image": image_url,
                "shape": "image",
                "title": Template(NODE_TEMPLATE).render(device=device_o),
            }
            # Set position
            if "positions" in details and str(device_id) in details["positions"]:
                nodes[device_id]["x"] = details["positions"][str(device_id)].get("x")
                nodes[device_id]["y"] = details["positions"][str(device_id)].get("y")

        # Create link
        cable_o = interface_o.cable
        cable_id = cable_o.pk
        if cable_id not in links:
            from_interface_o = cable_o.terminations.first().interface.first()
            to_interface_o = cable_o.terminations.last().interface.first()
            links[cable_id] = {
                "id": cable_id,
                "from": from_interface_o.device.pk,
                "from_label": from_interface_o.label,
                "to": to_interface_o.device.pk,
                "to_label": to_interface_o.label,
                "title": Template(CABLE_TEMPLATE).render(
                    from_interface=from_interface_o, to_interface=to_interface_o
                ),
            }

    return {
        "nodes": list(nodes.values()),
        "edges": list(links.values()),
    }


def get_l2_drawio_topology(queryset, diagram):
    """Create a L2 DrawIO topology data from an Interface queryset."""
    data = get_l2_topology_data(queryset, diagram.details)
    nodes = data.get("nodes")
    links = data.get("edges")

    # Transform node list into dict using id a key
    nodes_dict = {item["id"]: item for item in nodes}

    # Create diagram
    drawio_o = drawio_diagram()
    drawio_o.add_diagram("Page-1")

    for node in nodes:
        # Add node
        node_label = node.get("label")
        node_style = (
            node.get("role") if node.get("role") in DRAWIO_ROLE_MAP else "unknown"
        )
        drawio_o.add_node(
            id=node_label,
            url="Page-1",
            x_pos=node.get("x") if node.get("x") else None,
            y_pos=node.get("y") if node.get("x") else None,
            **DRAWIO_ROLE_MAP[node_style],
        )

    for link in links:
        # Add link (using node labels)
        drawio_o.add_link(
            nodes_dict[link["from"]]["label"],
            nodes_dict[link["to"]]["label"],
            src_label=link["from_label"],
            trgt_label=link["to_label"],
            link_id=link["id"],
        )

    return drawio_o.dump_xml()


def get_l3_topology_data(queryset, details):
    """Create a L3 vis.js topology data from an Interface queryset."""
    nodes = {}
    networks = {}
    links = {}

    for interface_o in queryset:
        for address_o in interface_o.ip_addresses.all():
            # We are using IPAddress's VRF. Interface's VRF is not relevant
            # Create device
            vrf_name = interface_o.vrf.name if interface_o.vrf else None
            device_o = interface_o.device
            device_name = device_o.name
            device_str = f"{device_name}:{vrf_name}" if vrf_name else device_name
            device_id = integer_hash(device_str)
            if device_id not in nodes:
                device_role = device_o.device_role.slug
                if device_role in [key for key, value in DeviceImageChoices()]:
                    image_url = f"/static/netdoc/img/{device_role}.png"
                else:
                    image_url = "/static/netdoc/img/unknown.png"
                nodes[device_id] = {
                    "id": device_id,
                    "label": device_str,
                    "role": device_role,
                    "image": image_url,
                    "shape": "image",
                    "title": Template(NODE_TEMPLATE).render(device=device_o),
                }
                # Set position
                if "positions" in details and str(device_id) in details["positions"]:
                    nodes[device_id]["x"] = details["positions"][str(device_id)].get(
                        "x"
                    )
                    nodes[device_id]["y"] = details["positions"][str(device_id)].get(
                        "y"
                    )

            # Create network
            network_o = ipaddress.ip_interface(address_o.address).network
            network_str = str(network_o.network_address.compressed)
            network_id = integer_hash(network_str)
            if network_id not in networks:
                networks[network_id] = {
                    "id": network_id,
                    "name": network_o.compressed,
                    "label": network_o.compressed,
                    "title": Template(NETWORK_TEMPLATE).render(
                        network=network_o, vrf=address_o.vrf
                    ),
                    "shape": "box",
                    "borderWidth": 2,
                    "color": {
                        "border": "#1b9bd0",
                        "background": "#36c6f4",
                        "highlight": {"border": "#1b9bd0", "background": "#36c6f4"},
                        "hover": {"border": "#1b9bd0", "background": "#36c6f4"},
                    },
                    "font": {
                        "color": "#000000",
                        "size": 14,
                    },
                }
                # Set position
                if "positions" in details and str(network_id) in details["positions"]:
                    networks[network_id]["x"] = details["positions"][
                        str(network_id)
                    ].get("x")
                    networks[network_id]["y"] = details["positions"][
                        str(network_id)
                    ].get("y")

            # Create link (Interface to Network)
            link_id = f"{device_id}-{network_id}"
            if link_id not in links:
                links[link_id] = {
                    "id": link_id,
                    "from": device_id,
                    "from_label": str(address_o.address),
                    "to": network_id,
                    "title": Template(NETWORK_TEMPLATE).render(
                        interface=interface_o, address=address_o
                    ),
                }

    return {
        "nodes": list(nodes.values()) + list(networks.values()),
        "edges": list(links.values()),
    }


def get_l3_drawio_topology(queryset, diagram):
    """Create a L3 DrawIO topology data from an Interface queryset."""
    data = get_l3_topology_data(queryset, diagram.details)
    nodes = data.get("nodes")
    links = data.get("edges")

    # Transform node list into dict using id a key
    nodes_dict = {item["id"]: item for item in nodes}

    # Create diagram
    drawio_o = drawio_diagram()
    drawio_o.add_diagram("Page-1")

    for node in nodes:
        # Add node
        node_label = node.get("label")
        node_style = (
            node.get("role") if node.get("role") in DRAWIO_ROLE_MAP else "unknown"
        )
        node_shape = node.get("shape")
        if node_shape == "image":
            # Device (using image/icon)
            drawio_o.add_node(
                id=node_label,
                url="Page-1",
                x_pos=node.get("x") if node.get("x") else None,
                y_pos=node.get("y") if node.get("x") else None,
                **DRAWIO_ROLE_MAP[node_style],
            )
        else:
            # Network (using a simple box)
            drawio_o.add_node(
                id=node_label,
                url="Page-1",
                width="80",
                height="30",
                x_pos=node.get("x") if node.get("x") else None,
                y_pos=node.get("y") if node.get("x") else None,
            )

    for link in links:
        # Add link (using node labels)
        drawio_o.add_link(
            nodes_dict[link["from"]]["label"],
            nodes_dict[link["to"]]["label"],
            src_label=link["from_label"],
            link_id=link["id"],
        )

    return drawio_o.dump_xml()



"""API View, called by API URLs."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from netbox.api.viewsets import NetBoxModelViewSet

from netdoc import models
from netdoc.api.serializers import (
    CredentialSerializer,
    DiscoverableSerializer,
    DiscoveryLogSerializer,
    DiagramSerializer,
)


class CredentialViewSet(NetBoxModelViewSet):
    """API View for add/edit Credential."""

    queryset = models.Credential.objects.prefetch_related("tags")
    serializer_class = CredentialSerializer


class DiagramViewSet(NetBoxModelViewSet):
    """API View for add/edit Diagram."""

    queryset = models.Diagram.objects.prefetch_related("tags")
    serializer_class = DiagramSerializer


class DiscoverableViewSet(NetBoxModelViewSet):
    """API View for add/edit Discoverable."""

    queryset = models.Discoverable.objects.prefetch_related("tags")
    serializer_class = DiscoverableSerializer


class DiscoveryLogViewSet(NetBoxModelViewSet):
    """API View for add/edit DiscoveryLog."""

    queryset = models.DiscoveryLog.objects.prefetch_related("tags")
    serializer_class = DiscoveryLogSerializer



"""Advanced filters."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from django import template
from django.urls import reverse

from utilities.utils import get_viewname

register = template.Library()


@register.inclusion_tag("netdoc/buttons/discover.html")
def discover_button(instance):
    """Add discover button.

    Used in templates/netdoc.
    """
    viewname = get_viewname(instance, "discover")
    url = reverse(viewname, kwargs={"pk": instance.pk})

    return {
        "url": url,
    }


@register.inclusion_tag("netdoc/buttons/export.html")
def export_button(instance):
    """Add export button.

    Used in templates/netdoc.
    """
    viewname = get_viewname(instance, "export")
    url = reverse(viewname, kwargs={"pk": instance.pk})

    return {
        "url": url,
    }



"""Schema validation for Cable/CableTermination."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from django.db.utils import IntegrityError
from django.contrib.contenttypes.models import ContentType

from dcim.models import Interface as Interface_model, Cable, CableTermination

from netdoc.schemas import interface

TERMINATION_TYPE = ContentType.objects.get(model="interface")


def get_schema():
    """Return the JSON schema to validate ArpTableEntry data."""
    return {
        "type": "object",
        "properties": {
            "left_interface_id": {
                "type": "integer",
                "enum": list(
                    Interface_model.objects.all().values_list("id", flat=True)
                ),
            },
            "right_interface_id": {
                "type": "integer",
                "enum": list(
                    Interface_model.objects.all().values_list("id", flat=True)
                ),
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new ArpTableEntry objects."""
    schema = get_schema()
    schema["required"] = [
        "left_interface_id",
        "right_interface_id",
    ]
    return schema


def sort_interfaces(**kwargs):
    """Neighborship must be sorted by Device name than Interface label.

    Ordered Interface obhects are returned.
    """
    # Get Interface objects
    left_interface_o = interface.get_list(id=kwargs.get("left_interface_id")).pop()
    right_interface_o = interface.get_list(id=kwargs.get("right_interface_id")).pop()

    if left_interface_o.device.name < right_interface_o.device.name:
        # Hostname comparison
        return left_interface_o, right_interface_o
    if left_interface_o.device.name > right_interface_o.device.name:
        # Hostname comparison
        return right_interface_o, left_interface_o
    if left_interface_o.name < right_interface_o.name:
        # Interface comparison
        return left_interface_o, right_interface_o
    # Interface names cannot be equal
    return right_interface_o, left_interface_o


def link(**kwargs):
    """Link two Interface objects.

    Both Interface objects must not be already linked.
    """
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())

    # Sort interfaces
    left_interface_o, right_interface_o = sort_interfaces(**kwargs)

    try:
        cable_o = Cable.objects.filter(terminations__interface=left_interface_o).get(
            terminations__interface=right_interface_o
        )
    except Cable.DoesNotExist:  # pylint: disable=no-member
        # Create Cable and CableTermination objects
        try:
            cable_o = Cable.objects.create()
            CableTermination.objects.create(
                termination=left_interface_o, cable=cable_o, cable_end="A"
            )
            CableTermination.objects.create(
                termination=right_interface_o, cable=cable_o, cable_end="B"
            )

            # Trigger Cable update
            cable_o._terminations_modified = True  # pylint: disable=protected-access
            cable_o.full_clean()
            cable_o.save()
        except IntegrityError as exc:
            # Crossed L2 connections with multiple devices (do we have a hub or a CDP/LLDP
            # forwarding device?)
            raise IntegrityError(
                f"Multiple neighbors on {left_interface_o.device}:{left_interface_o}"
                + f" or {right_interface_o.device}:{right_interface_o}"
            ) from exc


def unlink(**kwargs):
    """Unlink two Interface objects."""
    validate(kwargs, get_schema_create(), format_checker=FormatChecker())

    # Sort interfaces
    left_interface_o, right_interface_o = sort_interfaces(**kwargs)

    try:
        cable_o = Cable.objects.filter(terminations__interface=left_interface_o).get(
            terminations__interface=right_interface_o
        )
        cable_o.delete()
    except Cable.DoesNotExist:  # pylint: disable=no-member
        pass
    except IntegrityError as exc:
        raise IntegrityError(
            f"Multiple neighbors on {left_interface_o} or {right_interface_o}"
        ) from exc



"""Schema validation for Interface."""
__author__ = "Andrea Dainese"
__contact__ = "andrea@adainese.it"
__copyright__ = "Copyright 2022, Andrea Dainese"
__license__ = "GPLv3"

from jsonschema import validate, FormatChecker

from ipam.models import VLAN as VLAN_model, VRF as VRF_model
from dcim.models import Device as Device_model, Interface as Interface_model
from dcim.choices import (
    InterfaceTypeChoices,
    InterfaceDuplexChoices,
    InterfaceModeChoices,
)

from netdoc import utils
from netdoc.schemas import vlan, prefix, ipaddress


def get_interface_types():
    """Return Interface types."""
    types = []
    for key, value in InterfaceTypeChoices():  # pylint: disable=unused-variable
        for key1, value1 in value:  # pylint: disable=unused-variable
            types.append(key1)
    return types


def get_schema():
    """Return the JSON schema to validate Device data."""
    return {
        "type": "object",
        "properties": {
            "name": {
                "type": "string",
            },
            "device_id": {
                "type": "integer",
                "enum": list(Device_model.objects.all().values_list("id", flat=True)),
            },
            "type": {
                "type": "string",
                "enum": get_interface_types(),
            },
            "speed": {
                "type": "integer",
            },
            "duplex": {
                "type": "string",
                "enum": [key for key, value in InterfaceDuplexChoices()],
            },
            "vrf_id": {
                "type": "integer",
                "enum": list(VRF_model.objects.all().values_list("id", flat=True)),
            },
            "mac_address": {
                "type": "string",
            },
            "mtu": {
                "type": "integer",
            },
            "enabled": {
                "type": "boolean",
            },
            "parent_id": {
                "type": "integer",
                "enum": list(
                    Interface_model.objects.all().values_list("id", flat=True)
                ),
            },
            "bridge_id": {
                "type": "integer",
                "enum": list(
                    Interface_model.objects.all().values_list("id", flat=True)
                ),
            },
            "lag_id": {
                "type": "integer",
                "enum": list(
                    Interface_model.objects.all().values_list("id", flat=True)
                ),
            },
            "mode": {
                "type": "string",
                "enum": [key for key, value in InterfaceModeChoices()],
            },
            "untagged_vlan_id": {
                "type": "integer",
                "enum": list(VLAN_model.objects.all().values_list("vid", flat=True)),
            },
            "tagged_vlan_ids": {
                "type": "array",
                "item": {
                    "type": "integer",
                    "enum": list(
                        VLAN_model.objects.all().values_list("vid", flat=True)
                    ),
                },
            },
        },
    }


def get_schema_create():
    """Return the JSON schema to validate new Device objects."""
    schema = get_schema()
    schema["required"] = [
        "name",
        "type",
        "device_id",
    ]
    return schema


def create(int_type="other", **kwargs):
    """Create an Interface."""
    # The following are updated in update_mode
    if "mode" in kwargs:
        del kwargs["mode"]
    if "untagged_vlan_id" in kwargs:
        del kwargs["untagged_vlan_id"]
    if "tagged_vlan_ids" in kwargs:
        del kwargs["tagged_vlan_ids"]

    data = {
        **kwargs,
        "type": int_type,  # Default type is other
        "label": utils.normalize_interface_label(kwargs.get("name")),  # Set label
    }

    data = utils.delete_empty_keys(data)
    obj = utils.object_create(Interface_model, **data)
    return obj


def get(device_id, label):
    """Return an Interface."""
    obj = utils.object_get_or_none(Interface_model, device__id=device_id, label=label)
    return obj


def get_list(**kwargs):
    """Get a list of Interface objects."""
    validate(kwargs, get_schema(), format_checker=FormatChecker())
    result = utils.object_list(Interface_model, **kwargs)
    return result


def update(obj, **kwargs):
    """Update an Interface."""
    update_always = [
        "name",
        "description",
        "speed",
        "duplex",
        "vrf_id",
        "mac_address",
        "mtu",
        "enabled",
        "parent_id",
        "bridge_id",
        "lag_id",
        "parent_id",
        "bridge_id",
        "lag_id",
        "label",
    ]
    if obj.type == "other":
        # Updating Interface type only if current value is other
        # to avoid overwriting custom options.
        update_always.append("type")

    data = {
        **kwargs,
        "label": utils.normalize_interface_label(kwargs.get("name"))
        if kwargs.get("name")
        else None,  # Set label if name is updated
    }

    data = utils.delete_empty_keys(data)
    validate(data, get_schema(), format_checker=FormatChecker())
    kwargs_always = utils.filter_keys(data, update_always)
    obj = utils.object_update(obj, **kwargs_always)
    return obj


def update_mode(obj, mode=None, untagged_vlan=None, tagged_vlans=None):
    """Update an Interface mode.

    VLANs must exist.
    """
    if not tagged_vlans:
        tagged_vlans = []

    if mode == "tagged" and tagged_vlans and len(tagged_vlans) >= 4093:
        # Trunk with all VLANs (override mode)
        # In some switch the trunk is excluding native VLAN, so total is 4093.
        obj.mode = "tagged-all"
        obj.save()
    elif mode == "tagged" and tagged_vlans:
        # Trunk with some VLANs
        # Get current VLAN IDs and compare them with tagged_vlans
        for vid in tagged_vlans:
            # Add missing VLANs
            if not obj.tagged_vlans.filter(vid=vid):
                vlan_qs = vlan.get_list(vid=vid)
                if vlan_qs:
                    vlan_o = vlan_qs.pop()
                else:
                    # VLAN does not exist, creating one with default name
                    vlan_o = vlan.create(vid=vid, name=f"VLAN{vid:04d}")
                obj.tagged_vlans.add(vlan_o)

        for vlan_o in obj.tagged_vlans.all():
            # Remove unconfigured VLANs
            if vlan_o.vid not in tagged_vlans:
                obj.tagged_vlans.remove(vlan_o)
        obj.mode = mode
        obj.save()

    if untagged_vlan and mode in ["tagged", "access"]:
        # Get current VLAN IDs and compare them with untagged_vlan
        if not obj.untagged_vlan or obj.untagged_vlan != untagged_vlan:
            # A VLAN with a different VLAN ID is set
            vlan_qs = vlan.get_list(vid=untagged_vlan)
            if vlan_qs:
                vlan_o = vlan_qs.pop()
            else:
                # VLAN does not exist, creating one with default name
                vlan_o = vlan.create(vid=untagged_vlan, name=f"VLAN{untagged_vlan:04d}")
            data = {
                "mode": mode,
                "untagged_vlan": vlan_o,
            }
            obj = utils.object_update(obj, **data, force=False)

    return obj


def update_addresses(obj, ip_addresses=None):
    """Update Interface IP Addresses."""
    previous_ip_addresses_qs = obj.ip_addresses.all()
    previous_ip_addresses = [str(ip.address) for ip in previous_ip_addresses_qs]
    site_id = obj.device.site.id if obj.device.site else None
    vrf_id = obj.vrf.id if obj.vrf else None

    for ip_address in ip_addresses:
        if not ip_address:
            # Skip empty IP addresses
            continue
        # Get or create Prefix
        prefix_o = prefix.get(prefix=ip_address)
        if not prefix_o:
            prefix.create(prefix=ip_address, vrf_id=vrf_id, site_id=site_id)

        if ip_address not in previous_ip_addresses:
            # Get or create IPAddress
            ip_address_list = ipaddress.get_list(address=ip_address, vrf_id=vrf_id)
            if ip_address_list:
                ip_address_o = ip_address_list.pop()
            else:
                ip_address_o = ipaddress.create(address=ip_address, vrf_id=vrf_id)

            # Add missing IP address
            obj.ip_addresses.add(ip_address_o)

    for ip_address in previous_ip_addresses_qs:
        if str(ip_address) not in ip_addresses:
            # Get Interface IP Address objects
            ip_address_o = obj.ip_addresses.get(
                address__net_contains_or_equals=ip_address
            )
            # Remove unconfigured IP Address
            obj.ip_addresses.remove(ip_address_o)

    obj.save()
    return obj
